import os
import numpy as np
import pandas as pd
from tqdm import tqdm
from ase.optimize import FIRE
from ase.io import read, write
import matplotlib.pyplot as plt
from ase.build import add_adsorbate
from scipy.interpolate import griddata

# Ultility functions
def create_directory_if_not_exists(directory_path):
    if not os.path.exists(directory_path):
        try:
            os.makedirs(directory_path)
        except Exception as e:
            print(f"Failed to create directory '{directory_path}': {e}")
            raise

def pick_arrays(*list_other_info, 
                energies_list : list, 
                select : int):
    no_of_selection = select - 1
    sorted_array = np.array(energies_list).argsort()
    array_ranks = np.empty_like(sorted_array)
    array_ranks[sorted_array] = np.arange(len(energies_list))
    filtered_indices = [array_ranks[idx] for idx in range(len(array_ranks)) if array_ranks[idx] <= no_of_selection]
    unique_corresponding_list = [[lst[i] for i in filtered_indices] for lst in list_other_info]
    return [energies_list[i] for i in filtered_indices], unique_corresponding_list

def rank_lists(*all_list, 
               ref_list: list, 
               select=None):
    sorted_array = np.array(ref_list).argsort()
    sorted_array = sorted_array[:select]

    new_ref_list = [ref_list[idx] for idx in sorted_array]
    new_additional_list = [[lst[idx] for idx in sorted_array] for lst in all_list]
    print('Ranked the Lists')
    return new_additional_list, new_ref_list

def deduplicate_and_rank(*additional_lists, 
                         structures : list, 
                         energies_list : list, 
                         select : int, 
                         by_pos=True):
    no_of_selection = select - 1
    for lst in additional_lists:
        if len(structures) != len(lst):
            raise ValueError(f"structures other lists must have the same length, struct: {len(structures)} != lst: {len(lst)}")

    if by_pos:
        to_dedup = [tuple(map(tuple, structure.positions)) for structure in structures]
    else:
        to_dedup = [(ene,) for ene in energies_list]


    indexed_tuples = [(i, pos) for i, pos in enumerate(to_dedup)]

    seen = set()
    deduped_structures, dedup_ener_list =[], []
    unique_indices = []

    for index, element in indexed_tuples:
        if element not in seen:
            seen.add(element)
            deduped_structures.append(structures[index])
            dedup_ener_list.append(energies_list[index])
            unique_indices.append(index)

    new_additional_list = [np.array(lst)[np.sort(unique_indices)]
                           for lst in tqdm(additional_lists, desc='Removing duplicates', total= len(additional_lists))]

    sorted_array = np.array(energies_list).argsort()
    array_ranks = np.empty_like(sorted_array)
    array_ranks[sorted_array] = np.arange(len(energies_list))
    filtered_indices = [idx for idx in range(len(array_ranks)) if array_ranks[idx] <= no_of_selection]
    unique_corresponding_lists = []

    unique_corresponding_lists = [[lst[i] for i in filtered_indices] for lst in new_additional_list]

    return [deduped_structures[i] for i in filtered_indices], [energies_list[i] for i in filtered_indices], unique_corresponding_lists

def deduplicate(*additional_lists, 
                structures : list, 
                by_pos=True, 
                calculator):

    for lst in additional_lists:
        if len(structures) != len(lst):
            raise ValueError("structures other lists must have the same length")

    if by_pos:
        to_dedup = [tuple(map(tuple, structure.positions)) for structure in structures]
    else:
        pot_ene = []
        for structure in structures:
            structure.calc=calculator
            pot_ene.append(structure.get_potential_energy())
        to_dedup = [(ene,) for ene in pot_ene]


    indexed_tuples = [(i, pos) for i, pos in enumerate(to_dedup)]

    seen = set()
    deduped_structures =[]
    unique_indices = []

    for index, element in indexed_tuples:
        if element not in seen:
            seen.add(element)
            deduped_structures.append(structures[index])
            unique_indices.append(index)

    new_additional_list = [np.array(lst)[np.sort(unique_indices)]
                           for lst in tqdm(additional_lists, desc='Removing duplicates: ', total= len(additional_lists))]
    return deduped_structures, new_additional_list

def contour_plots(fig_save_path, 
                  list_energy, 
                  other_info, 
                  countours,
                  read_data, 
                  save_data, 
                  info_column, 
                  adsorbate, 
                  base, 
                  calculator
                  ):
    
    df = pd.read_csv(f'{read_data}/coordinates.csv')
    x_ax = df.iloc[:,0].values
    y_ax = df.iloc[:,1].values

    energies, (rest_info,) = pick_arrays(other_info, energies_list=list_energy, select=countours )

    create_directory_if_not_exists(save_data)
    
    df = pd.DataFrame(rest_info, columns=info_column)
    df.insert(0, 'Energy', energies)
    df.to_csv(f'{save_data}/single-point_screened_data.csv', index=False)

    create_directory_if_not_exists(fig_save_path)

    all_plot_energies, titles = [], []
    selected_structures = []
    #! Build molecule and create energy grid:
    for phi, theta, psi, x, y, z, file_name_prefix in rest_info:
        plot_energies = []
        for x, y in zip(x_ax, y_ax):
            titles.append(f'positions set: ({phi}-{theta}-{psi})-({x:.2f}-{y:.2f}-{z:.2f})')
            dummy = adsorbate.copy()
            dummy.euler_rotate(phi=phi, theta=theta, psi=psi, center= 'COM')
            slab = base.copy()
            add_adsorbate(slab, dummy, height=z, position=(x,y))
            slab.calc = calculator
            plot_energies.append(slab.get_potential_energy())
        
        for_selected_adsorbate = adsorbate.copy()
        for_selected_adsorbate.euler_rotate(phi=phi, theta=theta, psi=psi, center= 'COM')
        for_selected_base = base.copy()
        add_adsorbate(slab = for_selected_base, adsorbate=for_selected_adsorbate, position=(x,y), height=z)
        for_selected_base.calc = calculator
        selected_structures.append(for_selected_base)
        all_plot_energies.append(plot_energies)

    x_grid, y_grid =np.meshgrid(x_ax, y_ax)

    i = 0
    for energy_grid, title in tqdm(zip(all_plot_energies, titles), desc='Making Countours: ', total=len(all_plot_energies)):
        z = energy_grid
        zz = griddata((x_ax, y_ax), z, (x_grid, y_grid), method='cubic')
        # Plot the contour plot
        plt.figure(figsize=(10, 10))
        contour = plt.contourf(x_grid, y_grid, zz, levels=20, cmap='tab20b', alpha=1)
        plt.colorbar(contour, label='Energy')
        plt.scatter(x_ax, y_ax, c='green', s=50, alpha=0.4)
        plt.title(f'Contour Plot of phi, theta: {title}')
        plt.xlim(x_ax.min()-4, x_ax.max()+4)
        plt.ylim(y_ax.min()-4, y_ax.max()+4)
        plt.xlabel('X-axis Coordinates')
        plt.ylabel('Y-axis Coordinates')
        plt.savefig(f'{fig_save_path}/Plot-{i}.png')
        i+=1

    return selected_structures, rest_info

def mlip_optimizer(structures : list,
                   save_traj, 
                   calculator
                   ):
    optim_structs = []
    i = 0
    def optimize_structure(struct):
        nonlocal i
        i += 1
        struct.calc = calculator
        traj_file = f'{save_traj}/structure-{i}.vasp-xdatcar'
        print(f"Optimizing structure and saving trajectory to {traj_file}")
        FIRE(struct, trajectory=traj_file, logfile=None).run(steps=2000)
        optim_structs.append(struct)
        return struct.get_potential_energy()

    optim_structs_energies = list(map(optimize_structure, tqdm(structures, desc='Optimizing: ', total=  len(structures))))

    return optim_structs, optim_structs_energies


from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as f:
    description = f.read()

setup(
    name='pace',
    version='0.1.1',
    packages=find_packages(),
    install_requires=[
        # 'numpy>=1.24',
        # 'python>=3.10'

    ],
    long_description=description,
    long_description_content_type='text/markdown'

)



import itertools
import numpy as np
import pandas as pd
from tqdm import tqdm
from ase.optimize import FIRE
from ase.io import read, write
from ase.build import add_adsorbate
from ase.constraints import FixAtoms
from ase.io.trajectory import Trajectory


from pace.utilfuncs import *


class PACE:
    
    """
    PACE (Precise & Accelerated Configuration Evaluation) class for systematic
    screening of adsorbate positions and orientations over a surface.

    Attributes:
        base (ase.Atoms): The surface structure (slab).
        adsorbate (ase.Atoms): The adsorbate molecule to place on the surface.
        division (int): Number of subdivisions in x and y directions for lateral positions.
        z_levels (list of float): List of distances above the slab to try placing the adsorbate.
        arch (str): Folder name or identifier for model architecture and file paths.
        info_column (list of str): Metadata columns for generated conformations.
        mlip_filtered_structures_path (str): Directory to store filtered/optimized structures.
        screen_data_path (str): Directory to store screening data (CSV, figures).
        traj_path (str): Directory to store trajectory files.
    """
    def __init__(self, 
                 arch,  
                 base, 
                 adsorbate, 
                 division, 
                 z_levels):
        
        """
        Initialize the PACE class with necessary inputs.

        Args:
            arch (str): Identifier for model architecture and directory prefix.
            base (ase.Atoms): Base surface (slab) structure.
            adsorbate (ase.Atoms): Adsorbate molecule to screen.
            division (int): Number of divisions along the surface plane.
            z_levels (list of float): List of heights to test above the slab.
        """
        
        self.base = base
        self.adsorbate = adsorbate
        self.division = division
        self.z_levels = z_levels
        self.arch = arch
        self.info_column = ['phi', 'theta', 'psi', 'x', 'y', 'z', 'file_name_prefix']
        self.mlip_filtered_structures_path = 'mlip_optim_structs'
        self.screen_data_path = 'screen_data'
        self.traj_path = 'trajectory_data'

    def _setup(self):

        self.adsorbate.set_cell(self.base.cell)
        lattice_parameters = [self.base.cell[i][i] for i in range(3)]
        a, b, c = lattice_parameters

        cos = np.dot(self.base.cell[0], -self.base.cell[1]) / (np.linalg.norm(self.base.cell[0]) * np.linalg.norm(self.base.cell[1]))

        pos_x = [(d * a) / (self.division - 1) for d in range(self.division)]
        pos_y = [(d * b) / (self.division - 1) for d in range(self.division)]

        return cos, pos_x, pos_y
    
    def screen(self, 
               calculator, 
               fig_save_at : str,
               euler_angles=None, 
               mlip_optimization=20,
               make_countours=5):

        """
        Main function to screen adsorbate orientations and positions on a surface.

        This includes:
        - Optimization of base and adsorbate
        - Sampling of configurations (rotation, lateral and vertical translations)
        - Pre-screening using energy calculations
        - Contour plotting of energy surfaces
        - MLIP-based re-optimization of top candidates
        - Saving of optimized conformers

        Args:
            calculator (ase.calculators.Calculator): Calculator to compute energies.
            fig_save_at (str): Path to save generated contour plots.
            euler_angles (tuple or list of list, optional): Euler angles (phi, theta, psi) to rotate adsorbate.
            mlip_optimization (int): Number of top configurations to re-optimize with MLIP.
            make_countours (int): Number of top configurations to include in contour plots.

        Returns:
            dict: Dictionary with key 'screened_structures' and value as a list of ranked ase.Atoms objects.
        """

        model_arch_name = self.arch
        create_directory_if_not_exists(model_arch_name)

        cos, x_positions, y_positions = self._setup()
        self.adsorbate.calc = calculator; self.base.calc = calculator

        create_directory_if_not_exists(model_arch_name + '/' + self.mlip_filtered_structures_path)
        create_directory_if_not_exists(model_arch_name + '/' + self.traj_path)

        print("Starting optimization of adsorbate and base")
        
        FIRE(self.adsorbate, f'{model_arch_name}/{self.mlip_filtered_structures_path}/adsorb_opt.vasp-xdatcar', logfile=None).run(steps=2000)
        FIRE(self.base, f'{model_arch_name}/{self.mlip_filtered_structures_path}/base_opt.vasp-xdatcar', logfile=None).run(steps=2000)
        
        write(filename=model_arch_name+'/'+self.mlip_filtered_structures_path + f'base_optimised.vasp', images=self.base)
        write(filename=model_arch_name+'/'+self.mlip_filtered_structures_path + f'adsorbate_optimised.vasp', images=self.adsorbate)
        # self.base.set_constraint(FixAtoms(indices=[atom.index for atom in self.base]))

        if euler_angles != None:
            phis, thetas, psis = [euler_angles]*3

        else:
            phis, thetas, psis = [[0, -45, -90, -135, -180, -225, -270, -315]]*3

        x_positions_new = []
        coordinates_data = []
        
        for y in y_positions:
            for x in x_positions:
                x = x - (cos * np.sqrt(1 - np.square(cos)) * y)
                x_positions_new.append(x)
                coordinates_data.append((x,y))
                
        assign_angles = list(itertools.product(phis, thetas, psis))
        assign_positions = list(itertools.product(self.z_levels, y_positions, x_positions_new[:len(x_positions)]))

        adsorbate_conform, all_structures = [], []
        all_other_info = []

        create_directory_if_not_exists(model_arch_name + '/' + self.screen_data_path)
        pd.DataFrame(coordinates_data, columns=['x', 'y']).to_csv(f'{model_arch_name}/{self.screen_data_path}/coordinates.csv', index=False)

        prev_z = None
        counter = 0

        print('Initiating iterations')
        for phi, theta, psi in assign_angles:
            
            dummy = self.adsorbate.copy()
            dummy.euler_rotate(phi=phi, theta=theta, psi=psi, center='COP')
            dummy.positions -= dummy.get_center_of_mass()
            adsorbate_conform.append(dummy)

        # Remove duplicate adsorbate conformations
        adsorbate_conform, (assign_angles,) = deduplicate(assign_angles, structures=adsorbate_conform, calculator=calculator)


        global traj_files
        traj_files=[]
        for z, y, x in tqdm(assign_positions, desc='Making conformations: '):
            if z != prev_z:
                counter+=1
                prev_z = z

                try:
                    file_name_prefix = ''.join(str(z).split('.'))
                except:
                    file_name_prefix = str(z)
                zlvl_images = []

                traj_file = f'{model_arch_name}/{self.traj_path}/{file_name_prefix}-lvl.traj'
                traj_files.append(traj_file)

            for adsorbate, euler_angles in (zip(adsorbate_conform, assign_angles)):
                slab = self.base.copy()
                dummy_ads = adsorbate.copy()
                dummy_ads.translate([0, 0, slab.positions[:, 2].max() - dummy_ads.positions[:, 2].min()])
                dummy_ads.positions += [x,y,z]
                # adsorbate.translate([x, y, z])
                
                system = slab+dummy_ads
                # system.extend(dummy_ads)
                
                # add_adsorbate(system=system, adsorbate=adsorbate, height=z, position=(x,y))
                system.set_pbc([True, True, False])
                system.wrap()
                system.calc = calculator
                all_structures.append(system)
                with Trajectory(traj_file, mode='a') as traj:
                    traj.write(system)
                all_other_info.append((tuple(euler_angles) + ( x, y, z, file_name_prefix)))

        print(f"Generated trajectory files: {traj_files}")

        pd.DataFrame(all_other_info, columns=self.info_column).to_csv(f'{model_arch_name}/{self.screen_data_path}/conformations_generated.csv', index=False)
        if len(self.z_levels) > 1:
            write(f'{model_arch_name}/{self.traj_path}/to_screen.vasp-xdatcar', all_structures)
        print('TO SCREEN: ', len(all_structures))

        #! Calculate energies:
        def calc_energies(structure):
            structure.calc = calculator
            return structure.get_potential_energy()

        all_energies = list(map(calc_energies, tqdm(all_structures, desc='Pre-screening (SP): ', total=len(all_structures))))

        #! Make plots:
        print(f'Stage 2 complete \n Making Countour plots of top: {make_countours} ... \n')

        if make_countours > 0:                
            
            contour_filtered_structures, contour_filtered_info = contour_plots(fig_save_path=model_arch_name + '/' + fig_save_at, 
                                                                               adsorbate=self.adsorbate, 
                                                                               base=self.base,
                                                                               list_energy=all_energies, 
                                                                               other_info=all_other_info, 
                                                                               countours=make_countours,
                                                                               read_data=model_arch_name + '/' + self.screen_data_path, 
                                                                               save_data=model_arch_name + '/' + self.screen_data_path,
                                                                               calculator=calculator, 
                                                                               info_column=self.info_column
                                                                               )
            
        #! Deduplicate and rank:        
        (all_other_info, all_structures), all_energies = rank_lists(all_other_info, all_structures, 
                                                                    ref_list=all_energies, 
                                                                    select=mlip_optimization
                                                                    )
        
        # all_energies, all_structures, all_other_info = pick_arrays(all_structures, all_other_info, energies_list=all_energies, select=mlip_optimization)

        #! MLIP Optimization        
        print(f'Initiating optimisation of top {mlip_optimization} structures.')

        if mlip_optimization > 0:
            all_structures, all_energies = mlip_optimizer(structures=all_structures, 
                                                          save_traj=model_arch_name + '/' + self.traj_path, 
                                                          calculator=calculator
                                                          )
            
            df = pd.DataFrame(all_other_info, columns=self.info_column)
            df.insert(0, 'Energy', all_energies)
            df.to_csv(f'{model_arch_name}/{self.screen_data_path}/mlip_optimized_data.csv', index=False)
            # all_structures, all_other_info = deduplicate(all_other_info, structures=all_structures)
        else:
            all_structures = contour_filtered_structures
            all_other_info = contour_filtered_info

        
        #! saving all optimized conformations:
        sorted_array = np.array(all_energies).argsort(); array_ranks = np.empty_like(sorted_array)
        array_ranks[sorted_array] = np.arange(len(all_energies))
        ranked_structures = [all_structures[idx] for idx in sorted_array]
        for i in range(len(ranked_structures)):
            write(filename=model_arch_name+'/'+self.mlip_filtered_structures_path + "/" + f'structure_{i+1}.vasp', images=ranked_structures[i])
        # write(f'{model_arch_name}/{self.mlip_filtered_structures_path}/top_{mlip_optimization}_optimized_structures', all_structures, format='vasp-xdatcar')

        print(f"{'-x-o-'*20} \n SCREENING COMPLETE \n{'-x-o-'*20}")
        

        return dict({'screened_structures': ranked_structures})
